@@ -1,20 +1,23 @@
 """Automates poetry operations in the repo."""
 
-from pathlib import Path
 import re
-from typing import Set, Iterable, List, Union, Generator, Final
+from pathlib import Path
+from typing import Generator, Iterable, List, Set, Union
 
 import click
 from coveo_functools.finalizer import finalizer
-from coveo_stew.discovery import find_pyproject, discover_pyprojects
+from coveo_styles.styles import ExitWithFailure, echo, install_pretty_exception_hook
 from coveo_systools.filesystem import find_repo_root
-from coveo_styles.styles import echo, ExitWithFailure, install_pretty_exception_hook
 
-from coveo_stew.exceptions import CheckFailed, RequirementsOutdated, PythonProjectNotFound
+from coveo_stew.discovery import discover_pyprojects, find_pyproject
+from coveo_stew.exceptions import (
+    CheckFailed,
+    PythonProjectNotFound,
+    RequirementsOutdated,
+)
 from coveo_stew.offline_publish import offline_publish
 from coveo_stew.pydev import is_pydev_project, pull_and_write_dev_requirements
-from coveo_stew.stew import PythonProject, PythonEnvironment
-
+from coveo_stew.stew import PythonEnvironment, PythonProject
 
 _COMMANDS_THAT_SKIP_INTRO_EMOJIS = ["locate"]
 
@@ -69,7 +72,7 @@ def check_outdated(verbose: bool = False) -> None:
     try:
         for project in discover_pyprojects(verbose=verbose):
             echo.noise(project, item=True)
-            if not project.lock_path.exists() or project.lock_is_outdated:
+            if not project.lock_path.exists() or project.lock_is_outdated():
                 outdated.add(project.lock_path)
     except PythonProjectNotFound as exception:
         raise ExitWithFailure from exception
@@ -302,6 +305,7 @@ def refresh(project_name: str = None, exact_match: bool = False, verbose: bool =
 @click.option("--exact-match/--no-exact-match", default=False)
 @click.option("--fix/--no-fix", default=False)
 @click.option("--check", multiple=True, default=None)
+@click.option("--skip", multiple=True, default=None)
 @click.option("--verbose", is_flag=True, default=False)
 @click.option(
     "--quick",
@@ -309,13 +313,16 @@ def refresh(project_name: str = None, exact_match: bool = False, verbose: bool =
     default=False,
     help="Do not call 'poetry install --remove-untracked' before testing.",
 )
+@click.option("--parallel/--sequential", default=True)
 def ci(
     project_name: str = None,
     exact_match: bool = False,
     fix: bool = False,
     check: List[str] = None,
+    skip: List[str] = None,
     verbose: bool = False,
     quick: bool = False,
+    parallel: bool = True,
 ) -> None:
     failures = []
     try:
@@ -323,7 +330,9 @@ def ci(
             query=project_name, exact_match=exact_match, verbose=verbose
         ):
             echo.step(project.package.name, pad_after=False)
-            if not project.launch_continuous_integration(auto_fix=fix, checks=check, quick=quick):
+            if not project.launch_continuous_integration(
+                auto_fix=fix, checks=check, skips=skip, quick=quick, parallel=parallel
+            ):
                 failures.append(project)
     except PythonProjectNotFound as exception:
         raise ExitWithFailure from exception
@@ -332,5 +341,3 @@ def ci(
         raise ExitWithFailure(failures=failures) from CheckFailed(
             f"{len(failures)} project(s) failed ci steps."
         )
-
-    echo.success()
