@@ -1,34 +1,34 @@
 """Interact with python projects programmatically."""
 
-from contextlib import contextmanager
+import asyncio
 import os
-from pathlib import Path
 import re
 import shutil
-from shutil import rmtree
 import sys
-from typing import Generator, Optional, Any, List, Tuple, Iterator, Dict
-import warnings
+from contextlib import contextmanager
+from functools import cached_property
+from pathlib import Path
+from shutil import rmtree
+from typing import Any, Final, Generator, Iterator, List, Optional, Pattern, Tuple
 
 from coveo_functools.casing import flexfactory
 from coveo_itertools.lookups import dict_lookup
-from coveo_styles.styles import echo
-from coveo_systools.filesystem import find_repo_root, CannotFindRepoRoot
-from coveo_systools.subprocess import check_run, DetailedCalledProcessError
-from poetry.factory import Factory
+from coveo_systools.filesystem import CannotFindRepoRoot, find_repo_root
+from coveo_systools.subprocess import check_run
 
-from coveo_stew.ci.config import ContinuousIntegrationConfig
-from coveo_stew.ci.runner import RunnerStatus
-from coveo_stew.environment import PythonEnvironment, coveo_stew_environment, PythonTool
-from coveo_stew.exceptions import PythonProjectException, NotAPoetryProject
-from coveo_stew.metadata.stew_api import StewPackage
+from coveo_stew.environment import PythonEnvironment, PythonTool, find_python_tool
+from coveo_stew.exceptions import NotAPoetryProject, StewException
 from coveo_stew.metadata.poetry_api import PoetryAPI
-from coveo_stew.metadata.pyproject_api import PythonProjectAPI
 from coveo_stew.metadata.python_api import PythonFile
+from coveo_stew.metadata.stew_api import StewPackage
 from coveo_stew.utils import load_toml_from_path
 
+ENVIRONMENT_PATH_PATTERN: Final[Pattern] = re.compile(
+    r"^(?P<path>.+?)(?: (?P<activated>\(Activated\)))?$"
+)
 
-class PythonProject(PythonProjectAPI):
+
+class PythonProject:
     """Access the information within a pyproject.toml file and operate on them."""
 
     def __init__(self, project_path: Path, *, verbose: bool = False) -> None:
@@ -43,7 +43,7 @@ class PythonProject(PythonProjectAPI):
 
         try:
             self.package: PoetryAPI = flexfactory(
-                PoetryAPI, **dict_lookup(toml_content, "tool", "poetry"), _pyproject=self
+                PoetryAPI, **dict_lookup(toml_content, "tool", "poetry")
             )
         except KeyError as exception:
             raise NotAPoetryProject from exception
@@ -56,6 +56,8 @@ class PythonProject(PythonProjectAPI):
             _pyproject=self,
         )
 
+        from coveo_stew.ci.config import ContinuousIntegrationConfig  # circular import
+
         if self.options.pydev:
             # ensure no steps are repeated. pydev projects only receive basic poetry/lock checks
             self.ci: ContinuousIntegrationConfig = ContinuousIntegrationConfig(
@@ -68,64 +70,52 @@ class PythonProject(PythonProjectAPI):
                 _pyproject=self,
             )
 
-        # these are the actual poetry apis
-        self.poetry = Factory().create_poetry(self.project_path)
-
         try:
             repo_root: Optional[Path] = find_repo_root(self.project_path)
         except CannotFindRepoRoot:
             repo_root = None
 
         self.repo_root: Optional[Path] = repo_root
-        self._cached_environments: Optional[Dict[Path, PythonEnvironment]] = None
+
+    @cached_property
+    def _virtual_environments_cache(self) -> List[PythonEnvironment]:
+        cache: List[PythonEnvironment] = []
+
+        for path, activated in self._get_virtual_environment_paths():
+            environment = PythonEnvironment(path)
+            cache.append(environment)
+            environment.activated = activated
+
+        if cache and not any(environment.activated for environment in cache):
+            self.poetry_run("env", "use", cache[0].python_executable)
+            cache[0].activated = True
+
+        return cache
 
     def relative_path(self, path: Path) -> Path:
         """returns the relative path of a path vs the project folder."""
         return path.relative_to(self.project_path)
 
-    @classmethod
-    def find_pyproject(
-        cls, project_name: str, path: Path = None, *, verbose: bool = False
-    ) -> "PythonProject":
-        warnings.warn(
-            "This functionality moved to the `coveo_stew.discovery` module.", DeprecationWarning
-        )
-        from coveo_stew.discovery import find_pyproject
-
-        return find_pyproject(project_name, path, verbose=verbose)
-
-    @classmethod
-    def find_pyprojects(
-        cls,
-        path: Path = None,
-        *,
-        query: str = None,
-        exact_match: bool = False,
-        verbose: bool = False,
-    ) -> Generator["PythonProject", None, None]:
-        """Factory; scan a path (recursive) and return a PythonProject instance for each pyproject.toml
-
-        Parameters:
-            path: where to start looking for pyproject.toml files.
-            query: substring for package selection. '-' and '_' are equivalent.
-            exact_match: turns query into an exact match (except for - and _). Recommended use: CI scripts
-            verbose: output more details to command line
-        """
-        warnings.warn(
-            "This functionality moved to the `coveo_stew.discovery` module.", DeprecationWarning
-        )
-        from coveo_stew.discovery import discover_pyprojects
-
-        yield from discover_pyprojects(path, query=query, exact_match=exact_match, verbose=verbose)
-
-    @property
     def lock_is_outdated(self) -> bool:
         """True if the toml file has pending changes that were not applied to poetry.lock"""
         if not self.lock_path.exists():
             return False
-        return not self.poetry.locker.is_fresh()
 
-    @property
+        # yolo: use the dry run output to determine if the lock is too old
+        dry_run_output = self.poetry_run(
+            "install", "--remove-untracked", "--dry-run", capture_output=True
+        )
+
+        for sentence in (
+            "Warning: The lock file is not up to date",
+            "outdated dependencies",
+            "Run update to update them",
+        ):
+            if sentence.casefold() in dry_run_output.casefold():
+                return True
+
+        return False
+
     def activated_environment(self) -> Optional[PythonEnvironment]:
         """The environment activated for a project.
 
@@ -136,9 +126,11 @@ class PythonProject(PythonProjectAPI):
             - Use the context manager `self._activate_poetry_environment`
             - or use the `environment` argument of `self.poetry_run`
         """
+        if not self._virtual_environments_cache:
+            return None
+
         return next(
-            (environment for environment in self.virtual_environments() if environment.activated),
-            None,
+            environment for environment in self._virtual_environments_cache if environment.activated
         )
 
     def virtual_environments(
@@ -149,54 +141,31 @@ class PythonProject(PythonProjectAPI):
         create_default_if_missing: When no environments are found, install the environment using what poetry
             would use by default.
         """
-        reevaluate = self._cached_environments is None or all(
-            (
-                # we already have a cache
-                self._cached_environments is not None,
-                # but it's empty
-                not self._cached_environments,
-                # and we are supposed to create one
-                create_default_if_missing,
-            )
-        )
+        if not self._virtual_environments_cache and create_default_if_missing:
+            self._create_default_poetry_install()
 
-        if reevaluate:
-            self._refresh_virtual_environment_cache()
-            if not self._cached_environments and create_default_if_missing:
-                self.install()
-                self._refresh_virtual_environment_cache()
+        yield from self._virtual_environments_cache
 
-        yield from self._cached_environments.values()
+    def _create_default_poetry_install(self) -> PythonEnvironment:
+        """To be used only when no environments exist. Creates a default one by calling "poetry install"."""
+        self.poetry_run("install")
+        del self._virtual_environments_cache  # force cache refresh
+        activated_environment = self.activated_environment()
+        activated_environment.installed = True
+        activated_environment.cleaned = True
+        return activated_environment
 
-    def _refresh_virtual_environment_cache(self) -> None:
-        if self._cached_environments is None:
-            self._cached_environments = {}
-
-        for path in self._get_virtual_environment_paths():
-            if path not in self._cached_environments:
-                self._cached_environments[path] = PythonEnvironment(path)
-
-        # reprocess all environments to set the activated_environment one correctly
-        try:
-            activated_environment: Optional[PythonEnvironment] = PythonEnvironment(
-                self.poetry_run("env", "info", "--path", capture_output=True, breakout_of_venv=True)
-            )
-        except DetailedCalledProcessError as exception:
-            if exception.returncode != 1:
-                raise
-            activated_environment = None
-
-        for environment in self._cached_environments.values():
-            environment.activated = (
-                (environment == activated_environment) if activated_environment else False
-            )
-
-    def _get_virtual_environment_paths(self) -> Iterator[Path]:
+    def _get_virtual_environment_paths(self) -> Iterator[Tuple[Path, bool]]:
+        """Returns tuples of (path, activated)"""
         for str_path in self.poetry_run(
             "env", "list", "--full-path", capture_output=True, breakout_of_venv=True
         ).split("\n"):
-            if str_path.strip():
-                yield Path(str_path.replace("(Activated)", "").strip())
+            if (stripped := str_path.strip()) and (
+                match := re.fullmatch(ENVIRONMENT_PATH_PATTERN, stripped)
+            ):
+                yield Path(match.groupdict()["path"].strip()), bool(
+                    match.groupdict().get("activated")
+                )
 
     def current_environment_belongs_to_project(self) -> bool:
         """True if we're running from one of the project's virtual envs.
@@ -228,7 +197,7 @@ class PythonProject(PythonProjectAPI):
         wheel_match = wheel_pattern.search(poetry_output)
 
         if not wheel_match:
-            raise PythonProjectException(
+            raise StewException(
                 f"Unable able to find a wheel filename in poetry's output:\n{poetry_output}"
             )
 
@@ -251,51 +220,28 @@ class PythonProject(PythonProjectAPI):
 
         return target
 
+    def export(self) -> str:
+        """Generates the content of a `requirements.txt` file based on the lock."""
+        command = ["export"]
+        if self.options.build_without_hashes:
+            command.append("--without-hashes")
+
+        return self.poetry_run(*command, capture_output=True)
+
     def launch_continuous_integration(
-        self, auto_fix: bool = False, checks: List[str] = None, quick: bool = False
+        self,
+        auto_fix: bool = False,
+        checks: Optional[List[str]] = None,
+        skips: Optional[List[str]] = None,
+        quick: bool = False,
+        parallel: bool = True,
     ) -> bool:
         """Launch all continuous integration runners on the project."""
-        if self.ci.disabled:
-            return True
-
-        checks = [check.lower() for check in checks or []]
-        exceptions: List[DetailedCalledProcessError] = []
-        for environment in self.virtual_environments(create_default_if_missing=True):
-            if not quick:
-                self.install(environment=environment, remove_untracked=True)
-            for runner in self.ci.runners:
-                if checks and runner.name.lower() not in checks:
-                    continue
-
-                try:
-                    echo.normal(
-                        f"{runner} ({environment.pretty_python_version})", emoji="hourglass"
-                    )
-                    status = runner.launch(environment, auto_fix=auto_fix)
-                    if status is not RunnerStatus.Success:
-                        echo.warning(
-                            f"{self.package.name}: {runner} reported issues:",
-                            pad_before=False,
-                            pad_after=False,
-                        )
-                        runner.echo_last_failures()
-
-                except DetailedCalledProcessError as exception:
-                    echo.error(
-                        f"The ci runner {runner} failed to complete "
-                        f"due to an environment or configuration error."
-                    )
-                    exceptions.append(exception)
-
-        if exceptions:
-            if len(exceptions) > 1:
-                echo.warning(f"{len(exceptions)} exceptions found; raising first one.")
-            raise exceptions[0]
-
-        allowed_statuses: Tuple[RunnerStatus, ...] = (
-            (RunnerStatus.Success, RunnerStatus.NotRan) if checks else (RunnerStatus.Success,)
+        return asyncio.run(
+            self.ci.launch_continuous_integration(
+                auto_fix=auto_fix, checks=checks, skips=skips, quick=quick, parallel=parallel
+            )
         )
-        return all(runner.status in allowed_statuses for runner in self.ci.runners)
 
     def install(
         self,
@@ -307,8 +253,12 @@ class PythonProject(PythonProjectAPI):
         """
         Performs a 'poetry install --remove-untracked' on the project. If an environment is provided, target it.
         """
-        target_environment = environment or self.activated_environment
-        if target_environment and target_environment.installed:
+        target_environment = environment or self.activated_environment()
+        if not target_environment:
+            self._create_default_poetry_install()
+            return
+
+        if target_environment.installed:
             # this environment was already installed
             if not (remove_untracked and not target_environment.cleaned):
                 # return unless we are cleaning a non-cleaned environment
@@ -321,11 +271,8 @@ class PythonProject(PythonProjectAPI):
             command.append("--quiet")
 
         self.poetry_run(*command, environment=target_environment)
-
-        self._refresh_virtual_environment_cache()
-        affected_environment = target_environment or self.activated_environment
-        affected_environment.installed = True
-        affected_environment.cleaned |= remove_untracked
+        target_environment.installed = True
+        target_environment.cleaned |= remove_untracked
 
     def remove_egg_info(self) -> bool:
         """Removes the egg-info (editable project hook) from the folder. Returns True if we removed it."""
@@ -348,7 +295,7 @@ class PythonProject(PythonProjectAPI):
 
     def lock_if_needed(self) -> bool:
         """Lock if needed, return True if ran."""
-        if not self.lock_path.exists() or self.lock_is_outdated:
+        if not self.lock_path.exists() or self.lock_is_outdated():
             self.poetry_run("lock", breakout_of_venv=True)
             return True
         return False
@@ -360,23 +307,20 @@ class PythonProject(PythonProjectAPI):
         breakout_of_venv: bool = True,
         environment: PythonEnvironment = None,
     ) -> Optional[str]:
-        """internal run-a-poetry-command."""
-        # we use the poetry executable from our dependencies, not from the project's environment!
-        poetry_env = coveo_stew_environment
-        if not poetry_env.poetry_executable.exists():
-            raise PythonProjectException(
-                f"Poetry was not found; expected to be somewhere around {poetry_env.python_executable}?"
-            )
+        """
+        Internal run-a-poetry-command.
 
+        The `environment` param will make that environment active (e.g.: `poetry env use` called before).
+        """
         environment_variables = os.environ.copy()
         if breakout_of_venv:
             environment_variables.pop("VIRTUAL_ENV", None)
 
         with self._activate_poetry_environment(environment):
             return check_run(
-                *poetry_env.build_command(
-                    PythonTool.Poetry, *commands, "-vv" if self.verbose else ""
-                ),
+                *find_python_tool(PythonTool.Poetry, environment=environment),
+                *commands,
+                "-vv" if self.verbose else "",
                 working_directory=self.project_path,
                 capture_output=capture_output,
                 verbose=self.verbose,
@@ -393,7 +337,7 @@ class PythonProject(PythonProjectAPI):
             yield
             return
 
-        current_environment = self.activated_environment
+        current_environment = self.activated_environment()
         if current_environment == environment:
             yield
             return
